/* Includes ------------------------------------------------------------------*/
#include "UART.h"

#include <stdio.h>
#include <string.h>
#include "ring_buffer.h"

#include "sys/alt_irq.h"
#include "altera_avalon_uart_regs.h"
#include "system.h"

/* Private variables ---------------------------------------------------------*/


// UART transmit buffer descriptor
static RingBuffer UART_RingBuffer_Tx;
// UART transmit buffer memory pool
static char RingBufferData_Tx[128];

// UART receive buffer descriptor
static RingBuffer UART_RingBuffer_Rx;
// UART receive buffer memory pool
static char RingBufferData_Rx[128];

void (*UART_Callback_OnNewLine)(void);



bool UART_PutChar(char c){
	if(RingBuffer_PutChar(&UART_RingBuffer_Tx, c)){
		IOWR_ALTERA_AVALON_UART_CONTROL(UART0_BASE,
			IORD_ALTERA_AVALON_UART_CONTROL(UART0_BASE) | ALTERA_AVALON_UART_CONTROL_TRDY_MSK);
		return true;
	}else{
		return false;
	}
}


size_t UART_WriteData(const void *data, size_t dataSize){
	size_t i = 0;
	while(i<dataSize){
		if(UART_PutChar(((char*)data)[i])) i++;
	}
	return i;
}


size_t UART_WriteString(const char *string){
	return UART_WriteData(string, strlen(string));
}


bool UART_GetChar(char *c){
	return RingBuffer_GetChar(&UART_RingBuffer_Rx,c);
}


size_t UART_ReadData(char *data, size_t maxSize){
	size_t i = 0;
	for(; i < maxSize ; i++){
		if(!UART_GetChar(&data[i])) break;
	}
	return i;
}

bool UART_SetCallback_OnNewLine(void (*UART_Callback_OnNewLine_new)(void)){
	UART_Callback_OnNewLine = UART_Callback_OnNewLine_new;
	return true;
}

void uart0Interrupt(void* context){
	uint16_t status = IORD_ALTERA_AVALON_UART_STATUS(UART0_BASE);
	IOWR_ALTERA_AVALON_UART_STATUS(UART0_BASE, 0);

	if(status & ALTERA_AVALON_UART_STATUS_RRDY_MSK){
		char c = IORD_ALTERA_AVALON_UART_RXDATA(UART0_BASE);
		if(!RingBuffer_PutChar(&UART_RingBuffer_Rx, c)){
			//Buffer overrun
		}

		if(c == '\n'){
			if(UART_Callback_OnNewLine != NULL) UART_Callback_OnNewLine();
		}
	}

	if(status & ALTERA_AVALON_UART_STATUS_TRDY_MSK){
		char c;
		if(RingBuffer_GetChar(&UART_RingBuffer_Tx, &c)){
			IOWR_ALTERA_AVALON_UART_TXDATA(UART0_BASE, c);
		}else{
			IOWR_ALTERA_AVALON_UART_CONTROL(UART0_BASE,
				IORD_ALTERA_AVALON_UART_CONTROL(UART0_BASE) & ~ALTERA_AVALON_UART_CONTROL_TRDY_MSK);
		}
	}
}

/**
  * This function initialize ring buffers, UART device and enable receive interrupt.
  */
bool UART_Init(void){
	// initialize ring buffers
	RingBuffer_Init(&UART_RingBuffer_Tx, RingBufferData_Tx, sizeof(RingBufferData_Tx));
	RingBuffer_Init(&UART_RingBuffer_Rx, RingBufferData_Rx, sizeof(RingBufferData_Rx));
	UART_Callback_OnNewLine = NULL;
	// configure UART interrupt
	IOWR_ALTERA_AVALON_UART_CONTROL(UART0_BASE, ALTERA_AVALON_UART_CONTROL_RRDY_MSK);
	alt_ic_isr_register(UART0_IRQ_INTERRUPT_CONTROLLER_ID, UART0_IRQ, uart0Interrupt, NULL, NULL);
	return true;
}

void UART_SetBaud(uint32_t baud){
	uint16_t divisor = UART0_FREQ/baud;
	IOWR_ALTERA_AVALON_UART_DIVISOR(UART0_BASE, divisor);
}
